/*
 * main.c - FONTX2->BTRON font format converter
 * original code:SASANO Takayoshi
 * --- public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <basic.h>
#include <btron/btron.h>
#include <btron/font.h>
#include <errcode.h>
#include <tstring.h>
#include "fontx2import.h"

#define	KU_ELEMENTS	94
#define	TEN_ELEMENTS	94
#define	MAX_ELEMENTS	(KU_ELEMENTS * TEN_ELEMENTS)

/* $B6hE@$O(B1$B$+$i$G$O$J$/(B0$B$+$i?t$(;O$a$F$$$k(B */
#define	KuTen2JIS(ku, ten)	((((ku) + 0x21) << 8) | ((ten) + 0x21))

#define	Idx2Ku(idx)		((idx) / TEN_ELEMENTS)
#define	Idx2Ten(idx)		((idx) % TEN_ELEMENTS)
#define	Idx2JIS(idx)		KuTen2JIS(Idx2Ku((idx)), Idx2Ten((idx)))

LOCAL W		Map[MAX_ELEMENTS];

/* $B$3$$$D$i$O%0%m!<%P%kJQ?t$8$c$J$$J}$,NI$$$N$+$b$7$l$J$$$1$I(B */
LOCAL W		FdOut;
LOCAL UB	*Buf;
LOCAL FDEF	Arg;
    

/* $BJQ49:n6H(B */
LOCAL ERR	font_convert(W start_idx, W end_idx)
{
    WERR	err;
    W	i, x, y, start_ku, end_ku, charsize;
    UB	*data, *dummy;
    struct {
	FDEF	fdef;
	FIDATA	fidata;
    }	header;

    /* $BJ#?t$N6h$K$^$?$,$C$F$$$k>l9g!"6k7ANN0h$K$J$k$h$&Jd@5$9$k(B */
    start_ku = Idx2Ku(start_idx);
    end_ku   = Idx2Ku(end_idx);
    if(start_ku != end_ku) {
	start_idx = start_ku * TEN_ELEMENTS;
	end_idx   = (end_ku + 1) * TEN_ELEMENTS - 1;
    }

    printf("%04x->%04x\n", Idx2JIS(start_idx), Idx2JIS(end_idx));

    /* 1$BJ8;z$NI}!&9b$5!&%G!<%?%5%$%:$r<hF@$9$k(B */
    charsize = fontx2_getcharsize(Buf, &x, &y);

    /* $B%U%)%s%H$,B8:_$7$J$$>l9g$N%@%_!<$r:n@.(B */
    dummy = malloc(charsize);
    if(dummy == NULL) {
	err = ER_NOMEM;
	goto fin0;
    }
    memset(dummy, 0xff, charsize);

    /* $B%X%C%@$r:n@.$7!"=*C<%l%3!<%I$N0l$D<jA0$KDI2C$9$k(B */
    memset(&header, 0, sizeof(header));
    header.fdef.script     = 0x0021;
    header.fdef.fclass     = Arg.fclass;
    header.fdef.attr       = 
	FT_REGULAR | FT_MIDI | FT_RIGHT | FT_SOLID | FT_HDRAW | FT_FIXED;
    header.fdef.size       = y;
    header.fdef.width      = x;
    header.fdef.base       = y;
    header.fdef.leading    = 0;
    memcpy(header.fdef.name, Arg.name, sizeof(Arg.name));
    header.fdef.baseclass  = Arg.baseclass;
    memcpy(header.fdef.basename, Arg.basename, sizeof(Arg.basename));
    memcpy(header.fdef.fullname, Arg.fullname, sizeof(Arg.fullname));
    header.fdef.topcode    = Idx2JIS(start_idx);
    header.fdef.lastcode   = Idx2JIS(end_idx);
    header.fdef.sheight    = 0;
    header.fdef.swidth     = 0;
    /* header.fdef.rsv */
    header.fdef.imgform    = 0;
    header.fdef.widform    = 0;
    header.fdef.datasize   = 
	(end_idx - start_idx + 1) * charsize + sizeof(FIDATA);
    header.fdef.offimage   = sizeof(FDEF);
    header.fdef.offwidth   = 0;
    header.fdef.offnote    = 0;

    header.fidata.height   = y;
    header.fidata.width    = x;
    header.fidata.rowbytes = x / 8;
    header.fidata.margin   = 0;
    header.fidata.offset   = sizeof(FIDATA);

    err = apd_rec(FdOut, (UB *)&header, sizeof(header), 11, 0, 0);
    if(err < 0) {
	printf("apd_rec() %d\n", err);
	goto fin1;
    }

    /* $B=*C<%l%3!<%I$N0l$D<jA0$K0\F0$9$k(B */
    err = see_rec(FdOut, -1, -1, NULL);
    if(err < 0) {
	printf("see_rec() %d\n", err);
	goto fin1;
    }

    /* $BJ8;z?tJ,$N%U%)%s%H%G!<%?$r=PNO$9$k(B */
    for(i = start_idx; i <= end_idx; i++) {
	data = (Map[i] < 0) ? 
	    dummy : fontx2_getcharptr(Buf, Map[i], charsize);
	err = wri_rec(FdOut, sizeof(header) + charsize * (i - start_idx),
		      data, charsize, NULL, NULL, 0);
	if(err < 0) {
	    printf("wri_rec() %d\n", err);
	    goto fin1;
	}
    }

 fin1:
    free(dummy);
 fin0:
    return err;
}

/* $BB8:_$9$k%U%)%s%H$r8!:w$7!"JQ493+;O;X<($r=P$9(B */
LOCAL ERR	font_extract(VOID)
{
    ERR	err;
    W	i, current_state, prev_state, first_index, last_index;

    first_index = last_index = 0;
    current_state = prev_state = 0;

    for(i = 0; i < MAX_ELEMENTS; i++) {
	/* $B%U%)%s%H$NM-L5$r%A%'%C%/(B */
	current_state = (Map[i] < 0) ? 0 : 1;

	/* $B0l$D<jA0$N%U%)%s%H$N>uBV$H0[$J$C$F$$$k>l9g(B */
	if(current_state ^ prev_state) {
	    if(current_state) {
		/* $BL5"*M-(B: $B%U%)%s%H$N3+;O$r5-O?(B */
		first_index = i;
	    } else {
		/* $BM-"*L5(B: $B%U%)%s%H$N=*N;!"JQ49$r3+;O(B */
		err = font_convert(first_index, last_index);
		if(err < 0) return err;
	    }
	}

	/* $B%U%)%s%H>uBV$N99?7(B */
	last_index = i;
	prev_state = current_state;
    }

    return ER_OK;
}

/* $BJ8;z%3!<%I"*%U%)%s%HBP>HI=$N:n@.!JA43Q%U%)%s%H!K(B */
LOCAL ERR	makemap_kanji(VOID)
{
    W	i;

    for(i = 0; i < MAX_ELEMENTS; i++) {
	Map[i] = fontx2_getkanjifontindex(Buf, Idx2JIS(i));
    }

    return font_extract();
}

/* $BJ8;z%3!<%I"*%U%)%s%HBP>HI=$N:n@.!JH>3Q%U%)%s%H!K(B */
LOCAL ERR	makemap_ascii(VOID)
{
    printf("not implemented\n");
    return -1;
}

/* $BL>>N$J$I$N%Q%i%a!<%?$r@_Dj(B */
LOCAL VOID	set_argument(W argc, TC *argv[])
{
    W	len;
    TC	*dummy;

    memset(&Arg, 0, sizeof(Arg));

    len = tc_strlen(argv[3]) * sizeof(TC);
    memcpy(Arg.fullname, argv[3],
	   (len < sizeof(Arg.fullname)) ? len : sizeof(Arg.fullname));

    Arg.fclass = tc_strtol(argv[4], &dummy, 0);

    len = tc_strlen(argv[5]) * sizeof(TC);
    memcpy(Arg.name, argv[5],
	   (len < sizeof(Arg.name)) ? len : sizeof(Arg.name));

    if(argc > 6) {
	Arg.baseclass = tc_strtol(argv[6], &dummy, 0);

	len = tc_strlen(argv[7]) * sizeof(TC);
	memcpy(Arg.basename, argv[7],
	       (len < sizeof(Arg.basename)) ? len : sizeof(Arg.basename));
    }

    return;
}

/* $B%a%$%s(B: fontx2$B7A<0$N%G!<%?FI$_9~$_(B */
EXPORT ERR	main(W argc, TC *argv[])
{
    WERR	err = ER_PAR;
    W		fdin, size;
    LINK	linkin, linkout;

    /* $B%X%k%W$rI=<((B */
    if(argc != 6 && argc != 8) {
	printf("usage: fontx2btron [input_file] [output_file] [fullname] [fclass(hex)] [name] [[baseclass(hex)] [basename]]\n");
	goto fin0;
    }

    /* $B%Q%i%a!<%?$N@_Dj(B */
    set_argument(argc, argv);

    /* $BF~NO%U%!%$%k$X$N%j%s%/<h$j=P$7(B */
    err = get_lnk(argv[1], &linkin, F_NORM);
    if(err < 0) {
	printf("get_lnk(input_file) %d\n", err);
	goto fin0;
    }

    /* $BF~NO%U%!%$%k$N%*!<%W%s(B */
    err = opn_fil(&linkin, F_READ | F_EXCL, NULL);
    if(err < 0) {
	printf("opn_fil() %d\n", err);
	goto fin0;
    }
    fdin = err;

    /* $B%U%!%$%k%5%$%:$N<hF@(B */
    err = rea_rec(fdin, 0, NULL, 0, &size, NULL);
    if(err < 0) {
	printf("rea_rec() %d\n", err);
	goto fin1;
    }

    /* $B%a%b%j$N3NJ](B */
    Buf = malloc(size);
    if(Buf == NULL) goto fin1;

    /* $B%U%!%$%k$NFI$_9~$_(B */
    err = rea_rec(fdin, 0, Buf, size, NULL, NULL);
    if(err < 0) {
	printf("rea_rec() %d\n", err);
	goto fin2;
    }

    /* FONTX2$B7A<0$G$"$k$3$H$r3NG'(B */
    if(!fontx2_isfontx2(Buf)) {
	printf("not fontx2 format\n");
	goto fin2;
    }

    /* $B=PNO%U%!%$%k$X$N%j%s%/<h$j=P$7(B */
    err = get_lnk(NULL, &linkout, F_NORM);
    if(err < 0) {
	printf("get_lnk(output_file) %d\n", err);
	goto fin2;
    }

    /* $B=PNO%U%!%$%k$N:n@.(B */
    err = cre_fil(&linkout, argv[2], NULL, 0, F_FIX);
    if(err < 0) {
	printf("cre_fil() %d\n", err);
	goto fin2;
    }
    FdOut = err;

    /* $BA43Q!&H>3Q%U%)%s%H$K1~$8$FJQ493+;O(B */
    memset(Map, 0xff, sizeof(Map));
    if(fontx2_iskanjifont(Buf)) err = makemap_kanji();
    else if(fontx2_isasciifont(Buf)) err = makemap_ascii();
    else printf("invalid font type\n");

// fin3:
    cls_fil(FdOut);
 fin2:
    free(Buf);
 fin1:
    cls_fil(fdin);
 fin0:
    return err;
}
